// Read/Write request processors for AIOWDM driver

#include "stddcls.h"
#include "driver.h"

typedef NTKERNELAPI PHYSICAL_ADDRESS (*TPMmGetPhysicalAddress)(PVOID BaseAddress);

#pragma LOCKEDDATA

PDEVICE_EXTENSION GlobalExtension[MAX_SLOTS_OPEN];

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchCreate(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchCreate
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

	// Claim the remove lock in Win2K so that removal waits until the
	// handle closes. Don't do this in Win98, however, because this
	// device might be removed by surprise with handles open, whereupon
	// we'll deadlock in HandleRemoveDevice waiting for a close that
	// can never happen because we can't run the user-mode code that
	// would do the close.

	NTSTATUS status;
	if ( win98 || ! pdx->bRealDevice )
		status = STATUS_SUCCESS;
	else
		status = IoAcquireRemoveLock(&pdx->RemoveLock, stack->FileObject);

	if (NT_SUCCESS(status))
	{						// okay to open
		if (InterlockedIncrement(&pdx->handles) == 1)
		{					// first open handle
		}					// first open handle
		KdPrint(("Opened handle, %d handle%s open\n", pdx->handles, pdx->handles-1 ? "s" : ""));
	}					// okay to open
	return CompleteRequest(Irp, status, 0);
	}							// DispatchCreate

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

void ForgetIRP(PIRP IRP)
{
	PAGED_CODE();
	//This function is evil. However, if we complete the IRP, it'll ret into a thread that no
	//longer exists. This way, it'll remain pending indefinitely, consuming resources
	//including nonpaged memory... but the system won't bugcheck.

	//It would be really nice of us to delete the IRP without completing it, but we don't
	//know how, and furthermore don't believe there's a way(since it wouldn't be necessary
	//without this "ret into a thread that no longer exists" bug).
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchClose(PDEVICE_OBJECT fdo, PIRP Irp)
{							// DispatchClose
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	if (InterlockedDecrement(&pdx->handles) == 0)
	{						// no more open handles
		if (pdx->bRealDevice && pdx->InterruptRequested && pdx->HasPendingIrp)
		{
			pdx->InterruptRequested = FALSE;
			pdx->HasPendingIrp = FALSE;
			ForgetIRP(pdx->PendingIrp);
			pdx->PendingIrp = NULL;
			pdx->InterruptOccurred = FALSE;
		}
	}						// no more open handles
	KdPrint(("Closed handle, %d handle%s open\n", pdx->handles, pdx->handles-1 ? "s" : ""));

	// Release the remove lock to match the acquisition done in DispatchCreate

	if ( pdx->bRealDevice && ! win98 )
		IoReleaseRemoveLock(&pdx->RemoveLock, stack->FileObject);

	return CompleteRequest(Irp, STATUS_SUCCESS, 0);
}							// DispatchClose

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

VOID DMAUserReadsFreeAndComplete(TPDMAUserRead pFreeReads, NTSTATUS CompleteStatus)
{
	while ( pFreeReads )
	{
		TPDMAUserRead pFreeRead = pFreeReads;
		pFreeReads = pFreeReads->pNext;

		PIRP pPendingIRP = pFreeRead->pPendingIRP;
		ULONG BufferSize = pFreeRead->BufferSize;
		ExFreePool(pFreeRead);
		pFreeRead = NULL;

		MmUnlockPages(pPendingIRP->MdlAddress);

		pPendingIRP->IoStatus.Status = CompleteStatus;
		if ( NT_SUCCESS(CompleteStatus) )
			pPendingIRP->IoStatus.Information = BufferSize;
		else
			pPendingIRP->IoStatus.Information = 0
		;
		IoSetCancelRoutine(pPendingIRP, NULL);
		IoCompleteRequest(pPendingIRP, IO_SOUND_INCREMENT);
	}
}

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

VOID DpcForIsr(PKDPC Dpc, PDEVICE_OBJECT fdo, PIRP junk, PDEVICE_EXTENSION pdx)
{
	KIRQL OldIRQL;
	KeRaiseIrql(pdx->MyDIRQL, &OldIRQL);
	pdx->bNeedsDPC = FALSE;

	BOOLEAN bNeedsEnable = pdx->bNeedsEnable;
	pdx->bNeedsEnable = FALSE;

	TPDMAUserRead pFreeRead = pdx->pFirstFreeRead;
	pdx->pFirstFreeRead = NULL;
	pdx->pLastFreeRead = NULL;
	KeLowerIrql(OldIRQL);
	if ( bNeedsEnable ) DoIOAction(pdx, pdx->Enable);

	if (pdx->InterruptRequested && pdx->HasPendingIrp)
	{
		pdx->InterruptRequested = FALSE;
		pdx->InterruptOccurred = FALSE;
		pdx->HasPendingIrp = FALSE;
		pdx->PendingIrp->IoStatus.Status = STATUS_NO_MORE_FILES; //I wish STATUS_SUCCESS worked
		pdx->PendingIrp->IoStatus.Information = pdx->COSSize;
		IoSetCancelRoutine(pdx->PendingIrp, NULL);
		IoCompleteRequest(pdx->PendingIrp, IO_SOUND_INCREMENT);
	}
	else
	{
		pdx->InterruptOccurred = TRUE;
	}

	DMAUserReadsFreeAndComplete(pFreeRead, STATUS_TRANSLATION_COMPLETE); //I wish STATUS_SUCCESS worked
}

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

BOOLEAN NoSubISR(PDEVICE_EXTENSION pdx)
{
	return FALSE; //I don't represent a device that can interrupt, it wasn't me.
}

//

#pragma LOCKEDCODE

BOOLEAN BaseSubISR(PDEVICE_EXTENSION pdx)
{
	KdPrint((DRIVERNAME " - IRQ from %04LX?\n", pdx->DeviceID));
	UCHAR Data = QueryIOAction(pdx, pdx->IAm);

	if ( ! Data ) return FALSE; //then it wasn't me

	KdPrint((DRIVERNAME " - Yes!\n"));



	DoIOAction(pdx, pdx->Disable);
	pdx->bNeedsEnable = TRUE;

	if (pdx->InterruptRequested && pdx->HasPendingIrp)
		InCOSData(pdx, pdx->PendingIrp, pdx->COSSize)
	;

	DoIOAction(pdx, pdx->Clear);
	if ( ! pdx->bNeedsDPC )
	{
		pdx->bNeedsDPC = TRUE;
		//The DPC enables
		IoRequestDpc(pdx->DeviceObject, NULL, pdx);
	}

	return TRUE;
}

//

#pragma LOCKEDCODE

BOOLEAN DMASubISR(PDEVICE_EXTENSION pdx)
{
	KdPrint((DRIVERNAME " - IRQ from %04LX?\n", pdx->DeviceID));
	UCHAR Data = QueryIOAction(pdx, pdx->IAm);

	if ( ! Data ) return FALSE; //then it wasn't me

	KdPrint((DRIVERNAME " - Yes!\n"));

	LONG ClearType = -1;



	if ( pdx->bDMAGoing )
	{
		UCHAR IRQEvent = InBar8(pdx->MainBar, 0x2);
		if ( IRQEvent & (1 << 0) ) //FIFO AF.
		{
			BOOLEAN bDMAWrote = pdx->bDMAWrote;

			TPDMALink pOLink = &(pdx->DMALink[pdx->iDMALink]);
			if ( bDMAWrote ) pdx->iDMALink = pdx->iDMALink ^ 1;
			TPDMALink pNLink = &(pdx->DMALink[pdx->iDMALink]);

			OutBar32(pdx->StatusBar, 0x0, pNLink->LogicalAddress.LowPart);
			OutBar32(pdx->StatusBar, 0x4, pNLink->LogicalAddress.HighPart);
			OutBar32(pdx->StatusBar, 0x8, pdx->DMASizeUsed);
			OutBar32(pdx->StatusBar, 0xC, 0x4);

			pdx->bDMAWrote = TRUE;

			if ( bDMAWrote && pdx->pFirstRead )
			{
				TPDMAUserRead pRead = pdx->pFirstRead;
				pdx->pFirstRead = pdx->pFirstRead->pNext;
				if ( pdx->pFirstRead == NULL ) pdx->pLastRead = NULL;

				RtlCopyMemory(pRead->pBuffer, pOLink->pBuffer, pdx->DMASizeUsed);
				//`@ PULONG(pRead->pBuffer)[0] = IRQEvent | 0x80000000;

				pRead->pNext = NULL;
				if ( pdx->pLastFreeRead )
					pdx->pLastFreeRead->pNext = pRead;
				else
					pdx->pFirstFreeRead = pRead
				;
				pdx->pLastFreeRead = pRead;
			}
		}
		ClearType = IRQEvent;
	}
	else
	{
		DoIOAction(pdx, pdx->Disable);
		pdx->bNeedsEnable = TRUE;
	}

	if (pdx->InterruptRequested && pdx->HasPendingIrp)
		InCOSData(pdx, pdx->PendingIrp, pdx->COSSize)
	;

	if ( ClearType < 0 )
		DoIOAction(pdx, pdx->Clear);
	else
		OutBar8(pdx->MainBar, 0x2, UCHAR(ClearType));
	;
	if ( ! pdx->bNeedsDPC )
	{
		pdx->bNeedsDPC = TRUE;
		//The DPC enables
		IoRequestDpc(pdx->DeviceObject, NULL, pdx);
	}

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

BOOLEAN OnInterrupt(PKINTERRUPT InterruptObject, PVOID corrupt)
{
	KdPrint((DRIVERNAME " - IRQ...\n"));
	BOOLEAN Anybody = FALSE;

	for (int I = MAX_SLOTS_OPEN - 1; I >= 0 && GlobalExtension[I]; --I)
		if ( GlobalExtension[I] )
			if ( GlobalExtension[I]->SubISR(GlobalExtension[I]) )
				Anybody = TRUE
			//if ( DMASubISR(GlobalExtension[I]) )
			//	Anybody = TRUE
	;

	return Anybody;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

// Map port address, if on RISC platform.
NTSTATUS AlphaMapSpace(BOOLEAN bMap, PUCHAR * ToStoreMapped, PHYSICAL_ADDRESS Nonmapped, ULONG Size)
{
	if ( ! bMap )
		*ToStoreMapped = (PUCHAR) Nonmapped.QuadPart;
	else
	{
		*ToStoreMapped = (PUCHAR) MmMapIoSpace(Nonmapped, Size, MmNonCached);
		if ( ! *ToStoreMapped )
		{
			KdPrint((DRIVERNAME " - Unable to map port range %I64X, length %X\n", Nonmapped, Size));
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}
	return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID DeallocateDMALinks(PDEVICE_EXTENSION pdx)
{
	pdx->DMALinkCt = 0;
	while ( pdx->DMALinkDeallocate )
	{
		TPDMALink pLink;
		--pdx->DMALinkDeallocate;
		pLink = &(pdx->DMALink[pdx->DMALinkDeallocate]);
		//if ( pLink->pAdapter == NULL ) continue;
		if ( pLink->pBuffer == NULL ) continue;

		if ( pLink->pAdapter == NULL )
		{
			ExFreePool(pLink->pBuffer);
		}
		else
		{
			pLink->pAdapter->DmaOperations->FreeCommonBuffer(pLink->pAdapter, pLink->BufferSize, pLink->LogicalAddress, pLink->pBuffer, TRUE);
		}
		pLink->pBuffer = NULL;
		pLink->pAdapter = NULL;

		if ( KeGetCurrentIrql() < DISPATCH_LEVEL )
		{
			LOCAL_UNICODE_STRING(DebugData, 512);
			UStrAddStr(DebugData, L"Deallocated DMA[");
			UStrAddN(DebugData, pdx->DMALinkDeallocate);
			UStrAddStr(DebugData, L"] common buffer.");
			WriteRegStr(UDebugKey, iLogEvent++, DebugData);
		}
	}
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS StartDevice(PDEVICE_OBJECT fdo, PCM_PARTIAL_RESOURCE_LIST raw, PCM_PARTIAL_RESOURCE_LIST translated)
{							// StartDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	NTSTATUS status;

	status = GetDIDFor(pdx);
	if ( ! NT_SUCCESS(status) )
	{
		KdPrint((DRIVERNAME " - Could not get device ID!\n"));
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	// Identify the I/O resources we're supposed to use.

	ULONG I;
	ULONG vector;
	KIRQL irql = DISPATCH_LEVEL;
	KINTERRUPT_MODE mode;
	KAFFINITY affinity;
	BOOLEAN irqshare;
	BOOLEAN gotinterrupt = FALSE;
	//ULONG StatusSize, AuxSize;
	//BOOLEAN gotport = FALSE, GotStatus = FALSE, GotAux = FALSE;
	TBarData FoundBar[6];
	ULONG FoundBarCt = 0;

	if ( ! translated )
	{
		WriteRegStr(UDebugKey, iLogEvent++, L"Fatal Configuration Error: No resources assigned.");
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	PCM_PARTIAL_RESOURCE_DESCRIPTOR resource = translated->PartialDescriptors;
	ULONG nres = translated->Count;
	for ( I = 0; I < nres; ++I, ++resource )
	{						// for each resource
		switch (resource->Type)
		{					// switch on resource type

			case CmResourceTypePort: case CmResourceTypeMemory:
				FoundBar[FoundBarCt] = BarFromResource(*resource);
				++FoundBarCt;
			break;

			case CmResourceTypeInterrupt:
				irql = (KIRQL) resource->u.Interrupt.Level;
				vector = resource->u.Interrupt.Vector;
				affinity = resource->u.Interrupt.Affinity;
				mode = (resource->Flags == CM_RESOURCE_INTERRUPT_LATCHED) ?
					Latched : LevelSensitive;
				irqshare = resource->ShareDisposition == CmResourceShareShared;
				gotinterrupt = TRUE;
			break;

			case CmResourceTypeDevicePrivate:
				break;
			default:
				KdPrint((DRIVERNAME " - Unexpected I/O resource type %d\n", resource->Type));
			break;
		}					// switch on resource type
	}						// for each resource

	pdx->StatusBar = BlankBarData;
	pdx->MainBar = BlankBarData;
	pdx->AuxBar = BlankBarData;
	{
		LOCAL_UNICODE_STRING(DebugData, 256);
		UStrAddStr(DebugData, L"Starting device ID = ");
		UStrAdd16(DebugData, USHORT(pdx->DeviceID));
		UStrAddStr(DebugData, L".");
		WriteRegStr(UDebugKey, iLogEvent++, DebugData);
	}
	switch ( pdx->DeviceID )
	{
		case 0xC2E8: case 0xC2E9: case 0xC2EA:
		case 0x82E8: case 0x82E9: case 0x82EA:
		case 0xC258: case 0xC259: case 0xC25A:
		case 0x8258: case 0x8259: case 0x825A:
		case 0x1100:
			pdx->StatusBar = FoundBar[0];
			pdx->AuxBar = FoundBar[1];
			pdx->MainBar = FoundBar[2];

			//These are the DMA devices, so they use a different SubISR.
			pdx->SubISR = DMASubISR;
		break;
		default: switch ( FoundBarCt )
		{
			case 0:
				WriteRegStr(UDebugKey, iLogEvent++, L"Fatal Configuration Error: Didn't get needed PIO resources.");
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			break;
			case 1:
				pdx->MainBar = FoundBar[0];
			break;
			case 2:
				pdx->StatusBar = FoundBar[0];
				pdx->MainBar = FoundBar[1];
			break;
			case 3:
				pdx->StatusBar = FoundBar[0];
				pdx->AuxBar = FoundBar[1];
				pdx->MainBar = FoundBar[2];
			break;
			default: for ( I = 0; I < FoundBarCt; ++I )
			{
				if ( FoundBar[I].BarType == btMemory )
				{
					DelBar(FoundBar[I]);
				}
				else if ( pdx->MainBar.Base == NULL )
				{
					pdx->MainBar = FoundBar[I];
				}
				else if ( pdx->StatusBar.Base == NULL )
				{
					//The previous was the status bar, this is the main bar.
					pdx->StatusBar = pdx->MainBar;
					pdx->MainBar = FoundBar[I];
				}
				else if ( pdx->AuxBar.Base == NULL )
				{
					pdx->AuxBar = FoundBar[I];
				}
				else
				{
					DelBar(FoundBar[I]);
				}
			}
			break;
		}
		break;
	}

	status = GetClearIOFor(pdx);
	if ( ! NT_SUCCESS(status) )
	{
		WriteRegStr(UDebugKey, iLogEvent++, L"Fatal Configuration Error: Could not get clear I/O methods!");
		DelBar(pdx->StatusBar);
		DelBar(pdx->MainBar);
		DelBar(pdx->AuxBar);
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}
	if ( pdx->IAm.Operation == 'N' )
	{
		if ( pdx->StatusBar.Base == NULL )
		{
			//If there is no status bar, then this is a subtractive-decode device.
			pdx->IAm = BlankIOAction;
		}
		else if ( pdx->StatusBar.Size >= 0x100 )
		{
			pdx->IAm.BarID = 'S';
			pdx->IAm.Operation = 'R';
			pdx->IAm.Offset = 0x69;
			pdx->IAm.Data = 0x80;
		}
		else
		{
			pdx->IAm.BarID = 'S';
			pdx->IAm.Operation = 'R';
			pdx->IAm.Offset = 0x4C;
			pdx->IAm.Data = 0x04;
		}
	}

	DoIOAction(pdx, pdx->Disable);
	DoIOAction(pdx, pdx->Clear);
	//DoIOAction(pdx, pdx->Enable);

	//Attempts to correct .BIN error. Doesn't work.
	//if ( pdx->StatusBar.Size >= 0x100 ) OutBar32(pdx->StatusBar, 0x34, 0x00000050);

	pdx->DMALinkCt = 0;
	pdx->DMALinkDeallocate = 0;
	RtlZeroMemory(&(pdx->DMALink), sizeof(pdx->DMALink));
	for ( I = 0; I <= 1; ++I )
	{
		DEVICE_DESCRIPTION DD;
		ULONG MapRegisterCt;
		TPDMALink pLink = &(pdx->DMALink[I]);

		pLink->BufferSize = DMA_BUFFER_SIZE;

		RtlZeroMemory(&DD, sizeof(DD));
		DD.Version = DEVICE_DESCRIPTION_VERSION2;
		DD.Master = TRUE;
		DD.ScatterGather = TRUE;
		DD.DmaChannel = I;
		DD.InterfaceType = PCIBus;
		DD.MaximumLength = pLink->BufferSize;
		DD.Dma32BitAddresses = TRUE;
		DD.Dma64BitAddresses = TRUE;
		//DD.DmaAddressWidth = 64;
		DD.IgnoreCount = FALSE;
		MapRegisterCt = 0;
		pLink->pAdapter = IoGetDmaAdapter(pdx->Pdo, &DD, &MapRegisterCt);
		if ( pLink->pAdapter == NULL )
		{
			KdPrint((DRIVERNAME " - Failed to get DMA[%d] adapter.\n", I));
			if ( KeGetCurrentIrql() < DISPATCH_LEVEL )
			{
				LOCAL_UNICODE_STRING(DebugData, 256);
				UStrAddStr(DebugData, L"Failed to get DMA[");
				UStrAddN(DebugData, I);
				UStrAddStr(DebugData, L"] adapter.");
				WriteRegStr(UDebugKey, iLogEvent++, DebugData);
			}
			DelBar(pdx->StatusBar);
			DelBar(pdx->MainBar);
			DelBar(pdx->AuxBar);
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		++pdx->DMALinkDeallocate;
	}

	for ( I = 0; I < pdx->DMALinkDeallocate; ++I )
	{
		TPDMALink pLink = &(pdx->DMALink[I]);

		pLink->LogicalAddress.QuadPart = 0;
		pLink->pBuffer = PULONG(pLink->pAdapter->DmaOperations->AllocateCommonBuffer(pLink->pAdapter, pLink->BufferSize, &(pLink->LogicalAddress), TRUE));
		if ( pLink->pBuffer == NULL )
		{
			KdPrint((DRIVERNAME " - Failed to get DMA[%d] buffer.\n", I));
			if ( KeGetCurrentIrql() < DISPATCH_LEVEL )
			{
				LOCAL_UNICODE_STRING(DebugData, 256);
				UStrAddStr(DebugData, L"Failed to get DMA[");
				UStrAddN(DebugData, I);
				UStrAddStr(DebugData, L"] buffer.");
				WriteRegStr(UDebugKey, iLogEvent++, DebugData);
			}
			DeallocateDMALinks(pdx);
			DelBar(pdx->StatusBar);
			DelBar(pdx->MainBar);
			DelBar(pdx->AuxBar);
			return STATUS_INSUFFICIENT_RESOURCES;
		}

		if ( KeGetCurrentIrql() < DISPATCH_LEVEL )
		{
			LOCAL_UNICODE_STRING(DebugData, 512);
			UStrAddStr(DebugData, L"Allocated DMA[");
			UStrAddN(DebugData, I);
			UStrAddStr(DebugData, L"] a common buffer ");
			UStrAddN(DebugData, pLink->BufferSize);
			UStrAddStr(DebugData, L" B long at ");
			UStrAdd32(DebugData, pLink->LogicalAddress.HighPart);
			UStrAddStr(DebugData, L":");
			UStrAdd32(DebugData, pLink->LogicalAddress.LowPart);
			UStrAddStr(DebugData, L".");
			WriteRegStr(UDebugKey, iLogEvent++, DebugData);
		}

		RtlZeroMemory(pLink->pBuffer, pLink->BufferSize);

		++pdx->DMALinkCt;
	}

	pdx->GotIRQ = gotinterrupt;
	if ( pdx->GotIRQ )
	{
		status = IoConnectInterrupt(&pdx->InterruptObject, OnInterrupt, (PVOID) pdx, NULL, vector, irql, irql, mode, irqshare, affinity, FALSE);
		if (!NT_SUCCESS(status))
		{
			KdPrint((DRIVERNAME " - IoConnectInterrupt failed - %X\n", status));
			DeallocateDMALinks(pdx);
			DelBar(pdx->StatusBar);
			DelBar(pdx->MainBar);
			DelBar(pdx->AuxBar);
			return status;
		}
		if ( pdx->StatusBar.Size >= 0x100 ) OutBar8(pdx->StatusBar, 0x69, 0x09); // enable IRQs on PCIe bridge chip
		pdx->MyDIRQL = irql;
	}
	else
	{
		pdx->InterruptObject = NULL;
	}

	KdPrint((DRIVERNAME " - DeviceID = %04X, main base = %LX, status base = %LX\n", pdx->DeviceID, pdx->MainBar.Base, pdx->StatusBar.Base));

	return STATUS_SUCCESS;
}							// StartDevice

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID StopDevice(IN PDEVICE_OBJECT fdo, BOOLEAN oktouch /* = FALSE */)
{							// StopDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	if (pdx->InterruptObject)
	{						// disconnect interrupt

		DoIOAction(pdx, pdx->Disable);

		IoDisconnectInterrupt(pdx->InterruptObject);
		pdx->InterruptObject = NULL;
	}						// disconnect interrupt

	DeallocateDMALinks(pdx);

	DelBar(pdx->StatusBar);
	DelBar(pdx->MainBar);
	DelBar(pdx->AuxBar);
}							// StopDevice

